@@ -17,6 +17,8 @@ Lesser General Public License for more details.
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+
+@todo Feature request: The ability to go to relevant thing in web browser
 """
 
 from __future__ import with_statement
@@ -442,28 +444,34 @@ class Dialpad(object):
 
 		self._numberdisplay = widgetTree.get_widget("numberdisplay")
 		self._dialButton = widgetTree.get_widget("dial")
+		self._backButton = widgetTree.get_widget("back")
 		self._phonenumber = ""
 		self._prettynumber = ""
-		self._clearall_id = None
 
 		callbackMapping = {
 			"on_dial_clicked": self._on_dial_clicked,
 			"on_sms_clicked": self._on_sms_clicked,
 			"on_digit_clicked": self._on_digit_clicked,
 			"on_clear_number": self._on_clear_number,
-			"on_back_clicked": self._on_backspace,
-			"on_back_pressed": self._on_back_pressed,
-			"on_back_released": self._on_back_released,
 		}
 		widgetTree.signal_autoconnect(callbackMapping)
 
+		self._originalLabel = self._backButton.get_label()
+		self._backTapHandler = gtk_toolbox.TapOrHold(self._backButton)
+		self._backTapHandler.on_tap = self._on_backspace
+		self._backTapHandler.on_hold = self._on_clearall
+		self._backTapHandler.on_holding = self._set_clear_button
+		self._backTapHandler.on_cancel = self._reset_back_button
+
 		self._window = gtk_toolbox.find_parent_window(self._numberdisplay)
 
 	def enable(self):
 		self._dialButton.grab_focus()
+		self._backTapHandler.enable()
 
 	def disable(self):
-		pass
+		self._reset_back_button()
+		self._backTapHandler.disable()
 
 	def number_selected(self, action, number, message):
 		"""
@@ -483,7 +491,7 @@ class Dialpad(object):
 			self._prettynumber = make_pretty(self._phonenumber)
 			self._numberdisplay.set_label("<span size='30000' weight='bold'>%s</span>" % (self._prettynumber))
 		except TypeError, e:
-			self._errorDisplay.push_exception(e)
+			self._errorDisplay.push_exception()
 
 	def clear(self):
 		self.set_number("")
@@ -526,49 +534,104 @@ class Dialpad(object):
 	def _on_digit_clicked(self, widget):
 		self.set_number(self._phonenumber + widget.get_name()[-1])
 
-	def _on_backspace(self, widget):
-		self.set_number(self._phonenumber[:-1])
+	def _on_backspace(self, taps):
+		self.set_number(self._phonenumber[:-taps])
+		self._reset_back_button()
 
-	def _on_clearall(self):
+	def _on_clearall(self, taps):
 		self.clear()
+		self._reset_back_button()
 		return False
 
-	def _on_back_pressed(self, widget):
-		self._clearall_id = gobject.timeout_add(1000, self._on_clearall)
+	def _set_clear_button(self):
+		self._backButton.set_label("gtk-clear")
 
-	def _on_back_released(self, widget):
-		if self._clearall_id is not None:
-			gobject.source_remove(self._clearall_id)
-		self._clearall_id = None
+	def _reset_back_button(self):
+		self._backButton.set_label(self._originalLabel)
 
 
 class AccountInfo(object):
 
-	def __init__(self, widgetTree, backend, errorDisplay):
+	def __init__(self, widgetTree, backend, alarmHandler, errorDisplay):
 		self._errorDisplay = errorDisplay
 		self._backend = backend
 		self._isPopulated = False
+		self._alarmHandler = alarmHandler
+		self._notifyOnMissed = False
+		self._notifyOnVoicemail = False
+		self._notifyOnSms = False
 
 		self._callbackList = gtk.ListStore(gobject.TYPE_STRING)
 		self._accountViewNumberDisplay = widgetTree.get_widget("gcnumber_display")
 		self._callbackCombo = widgetTree.get_widget("callbackcombo")
 		self._onCallbackentryChangedId = 0
 
+		self._notifyCheckbox = widgetTree.get_widget("notifyCheckbox")
+		self._minutesEntry = widgetTree.get_widget("minutesEntry")
+		self._missedCheckbox = widgetTree.get_widget("missedCheckbox")
+		self._voicemailCheckbox = widgetTree.get_widget("voicemailCheckbox")
+		self._smsCheckbox = widgetTree.get_widget("smsCheckbox")
+		self._onNotifyToggled = 0
+		self._onMinutesChanged = 0
+		self._onMissedToggled = 0
+		self._onVoicemailToggled = 0
+		self._onSmsToggled = 0
+
 		self._defaultCallback = ""
 
 	def enable(self):
 		assert self._backend.is_authed(), "Attempting to enable backend while not logged in"
+
 		self._accountViewNumberDisplay.set_use_markup(True)
 		self.set_account_number("")
+
 		self._callbackList.clear()
 		self._onCallbackentryChangedId = self._callbackCombo.get_child().connect("changed", self._on_callbackentry_changed)
+
+		if self._alarmHandler is not None:
+			self._notifyCheckbox.set_active(self._alarmHandler.isEnabled)
+			self._minutesEntry.set_value(self._alarmHandler.recurrence)
+			self._missedCheckbox.set_active(self._notifyOnMissed)
+			self._voicemailCheckbox.set_active(self._notifyOnVoicemail)
+			self._smsCheckbox.set_active(self._notifyOnSms)
+
+			self._onNotifyToggled = self._notifyCheckbox.connect("toggled", self._on_notify_toggled)
+			self._onMinutesChanged = self._minutesEntry.connect("value-changed", self._on_minutes_changed)
+			self._onMissedToggled = self._missedCheckbox.connect("toggled", self._on_missed_toggled)
+			self._onVoicemailToggled = self._voicemailCheckbox.connect("toggled", self._on_voicemail_toggled)
+			self._onSmsToggled = self._smsCheckbox.connect("toggled", self._on_sms_toggled)
+		else:
+			self._notifyCheckbox.set_sensitive(False)
+			self._minutesEntry.set_sensitive(False)
+			self._missedCheckbox.set_sensitive(False)
+			self._voicemailCheckbox.set_sensitive(False)
+			self._smsCheckbox.set_sensitive(False)
+
 		self.update(force=True)
 
 	def disable(self):
 		self._callbackCombo.get_child().disconnect(self._onCallbackentryChangedId)
+		self._onCallbackentryChangedId = 0
+
+		if self._alarmHandler is not None:
+			self._notifyCheckbox.disconnect(self._onNotifyToggled)
+			self._minutesEntry.disconnect(self._onMinutesChanged)
+			self._missedCheckbox.disconnect(self._onNotifyToggled)
+			self._voicemailCheckbox.disconnect(self._onNotifyToggled)
+			self._smsCheckbox.disconnect(self._onNotifyToggled)
+			self._onNotifyToggled = 0
+			self._onMinutesChanged = 0
+			self._onMissedToggled = 0
+			self._onVoicemailToggled = 0
+			self._onSmsToggled = 0
+		else:
+			self._notifyCheckbox.set_sensitive(True)
+			self._minutesEntry.set_sensitive(True)
+			self._missedCheckbox.set_sensitive(True)
+			self._voicemailCheckbox.set_sensitive(True)
+			self._smsCheckbox.set_sensitive(True)
 
 		self.clear()
-
 		self._callbackList.clear()
 
 	def get_selected_callback_number(self):
@@ -582,21 +645,28 @@ class AccountInfo(object):
 
 	def update(self, force = False):
 		if not force and self._isPopulated:
-			return
+			return False
 		self._populate_callback_combo()
 		self.set_account_number(self._backend.get_account_number())
+		return True
 
 	def clear(self):
 		self._callbackCombo.get_child().set_text("")
 		self.set_account_number("")
 		self._isPopulated = False
 
+	def save_everything(self):
+		raise NotImplementedError
+
 	@staticmethod
 	def name():
 		return "Account Info"
 
 	def load_settings(self, config, section):
 		self._defaultCallback = config.get(section, "callback")
+		self._notifyOnMissed = config.getboolean(section, "notifyOnMissed")
+		self._notifyOnVoicemail = config.getboolean(section, "notifyOnVoicemail")
+		self._notifyOnSms = config.getboolean(section, "notifyOnSms")
 
 	def save_settings(self, config, section):
 		"""
@@ -604,6 +674,9 @@ class AccountInfo(object):
 		"""
 		callback = self.get_selected_callback_number()
 		config.set(section, "callback", callback)
+		config.set(section, "notifyOnMissed", repr(self._notifyOnMissed))
+		config.set(section, "notifyOnVoicemail", repr(self._notifyOnVoicemail))
+		config.set(section, "notifyOnSms", repr(self._notifyOnSms))
 
 	def _populate_callback_combo(self):
 		self._isPopulated = True
@@ -611,7 +684,7 @@ class AccountInfo(object):
 		try:
 			callbackNumbers = self._backend.get_callback_numbers()
 		except StandardError, e:
-			self._errorDisplay.push_exception(e)
+			self._errorDisplay.push_exception()
 			self._isPopulated = False
 			return
 
@@ -648,13 +721,44 @@ class AccountInfo(object):
 					UserWarning, 2
 				)
 		except StandardError, e:
-			self._errorDisplay.push_exception(e)
+			self._errorDisplay.push_exception()
+
+	def _update_alarm_settings(self):
+		try:
+			isEnabled = self._notifyCheckbox.get_active()
+			recurrence = self._minutesEntry.get_value_as_int()
+			if isEnabled != self._alarmHandler.isEnabled or recurrence != self._alarmHandler.recurrence:
+				self._alarmHandler.apply_settings(isEnabled, recurrence)
+		finally:
+			self.save_everything()
+			self._notifyCheckbox.set_active(self._alarmHandler.isEnabled)
+			self._minutesEntry.set_value(self._alarmHandler.recurrence)
 
 	def _on_callbackentry_changed(self, *args):
 		text = self.get_selected_callback_number()
 		number = make_ugly(text)
 		self._set_callback_number(number)
 
+		self.save_everything()
+
+	def _on_notify_toggled(self, *args):
+		self._update_alarm_settings()
+
+	def _on_minutes_changed(self, *args):
+		self._update_alarm_settings()
+
+	def _on_missed_toggled(self, *args):
+		self._notifyOnMissed = self._missedCheckbox.get_active()
+		self.save_everything()
+
+	def _on_voicemail_toggled(self, *args):
+		self._notifyOnVoicemail = self._voicemailCheckbox.get_active()
+		self.save_everything()
+
+	def _on_sms_toggled(self, *args):
+		self._notifyOnSms = self._smsCheckbox.get_active()
+		self.save_everything()
+
 
 class RecentCallsView(object):
 
@@ -737,8 +841,9 @@ class RecentCallsView(object):
 
 	def update(self, force = False):
 		if not force and self._isPopulated:
-			return
+			return False
 		self._updateSink.send(())
+		return True
 
 	def clear(self):
 		self._isPopulated = False
@@ -764,7 +869,7 @@ class RecentCallsView(object):
 		try:
 			recentItems = self._backend.get_recent()
 		except StandardError, e:
-			self._errorDisplay.push_exception_with_lock(e)
+			self._errorDisplay.push_exception_with_lock()
 			self._isPopulated = False
 			recentItems = []
 
@@ -884,8 +989,9 @@ class MessagesView(object):
 
 	def update(self, force = False):
 		if not force and self._isPopulated:
-			return
+			return False
 		self._updateSink.send(())
+		return True
 
 	def clear(self):
 		self._isPopulated = False
@@ -911,7 +1017,7 @@ class MessagesView(object):
 		try:
 			messageItems = self._backend.get_messages()
 		except StandardError, e:
-			self._errorDisplay.push_exception_with_lock(e)
+			self._errorDisplay.push_exception_with_lock()
 			self._isPopulated = False
 			messageItems = []
 
@@ -1050,8 +1156,9 @@ class ContactsView(object):
 
 	def update(self, force = False):
 		if not force and self._isPopulated:
-			return
+			return False
 		self._updateSink.send(())
+		return True
 
 	def clear(self):
 		self._isPopulated = False
@@ -1094,7 +1201,7 @@ class ContactsView(object):
 			except StandardError, e:
 				contacts = []
 				self._isPopulated = False
-				self._errorDisplay.push_exception_with_lock(e)
+				self._errorDisplay.push_exception_with_lock()
 			for contactId, contactName in contacts:
 				contactType = (addressBook.contact_source_short_name(contactId), )
 				self._contactsmodel.append(contactType + (contactName, "", contactId) + ("", ))
@@ -1124,7 +1231,7 @@ class ContactsView(object):
 			contactDetails = self._addressBook.get_contact_details(contactId)
 		except StandardError, e:
 			contactDetails = []
-			self._errorDisplay.push_exception(e)
+			self._errorDisplay.push_exception()
 		contactPhoneNumbers = [phoneNumber for phoneNumber in contactDetails]
 
 		if len(contactPhoneNumbers) == 0:
